"""Functions for predicting protein coding and tRNA/tmRNA genes on
bacterial contigs.
"""

from Bio import SeqIO
from Bio.SeqFeature import SeqFeature, FeatureLocation

from depht.data import GLOBAL_VARIABLES, PARAMETERS
from depht.functions.fasta import parse_fasta
from depht.functions.subprocess import run_command

CODING_FEATURE_TYPES = GLOBAL_VARIABLES["sequences"]["feature_types"]
DEFAULT_PRODUCT = GLOBAL_VARIABLES["sequences"]["default_product"]

# Don't annotate short contigs
MIN_LENGTH = PARAMETERS["annotation"]["min_length"]
# Medium-length contigs -> use metagenomic mode
META_LENGTH = PARAMETERS["annotation"]["meta_length"]


def run_prodigal(infile, outfile, meta=False):
    """Run Prodigal on the indicated input file, in metagenomic mode
    if `meta` == True, with predicted CDS features written to the
    indicated output file.

    :param infile: the input file with nucleotide sequence to annotate
    :type infile: pathlib.Path
    :param outfile: the output file where predicted genes should go
    :type outfile: pathlib.Path
    :param meta: run in metagenomic mode?
    :type meta: bool
    :return: outfile
    """
    command = f"prodigal -i {infile} -a {outfile} -n -c"
    if meta:
        command += " -p meta"

    try:
        stdout, stderr = run_command(command)
    except FileNotFoundError:
        # Prodigal not found
        raise RuntimeError("Unable to locate Prodigal")

    return outfile


def parse_prodigal(outfile):
    """
    Parses Prodigal output file into a list of BioPython SeqFeatures.

    :param outfile: the path to the output file written by Prodigal
    :type outfile: pathlib.Path
    :return: features
    """
    features = list()

    headers, sequences = parse_fasta(outfile)
    for header, sequence in zip(headers, sequences):
        header = header.split(" # ")
        start, end, strand = int(header[1]), int(header[2]), int(header[3])
        notes = header[-1].split(";")
        motif = notes[-3].split("=")[-1]
        spacer = notes[-2].split("=")[-1]

        ftr = SeqFeature(location=FeatureLocation(start - 1, end),
                         type="CDS", strand=strand)
        ftr.qualifiers["gene"] = [""]
        ftr.qualifiers["locus_tag"] = [""]
        ftr.qualifiers["note"] = [f"rbs_motif: {motif}; rbs_spacer: {spacer}"]
        ftr.qualifiers["transl_table"] = [11]
        ftr.qualifiers["product"] = [DEFAULT_PRODUCT]
        ftr.qualifiers["translation"] = [sequence.rstrip("*")]

        features.append(ftr)

    return features


def run_aragorn(infile, outfile):
    """Run Aragorn on the indicated input file, with predicted tRNA
    and tmRNA features written to the indicated output file.

    :param infile: the input file to be used by Aragorn
    :type infile: pathlib.Path
    :param outfile: the output file to be written by Aragorn
    :type outfile: pathlib.Path
    :return: outfile
    """
    command = f"aragorn -gcbact -l -d -wa -o {outfile} {infile}"

    try:
        stdout, stderr = run_command(command)
    except FileNotFoundError:
        # Aragorn not found
        raise RuntimeError("Unable to locate Aragorn")

    return outfile


def parse_aragorn(outfile):
    """
    Parses Aragorn output file into a list of BioPython SeqFeatures.

    :param outfile: the path to the output file written by Aragorn
    :type outfile: pathlib.Path
    :return: features
    """
    features = list()

    aragorn_reader = open(outfile, "r")

    # Skip the 2 header lines
    for _ in range(2):
        next(aragorn_reader)

    # Now parse the remaining lines into Bio.SeqFeature.SeqFeature objects
    for row in aragorn_reader:
        # Tokenize the line; skip the tRNA number, which is meaningless
        row = row.rstrip().split()[1:]

        # Get coordinates
        if row[1].startswith("c"):
            strand = -1  # reverse oriented
            coords = row[1][2:-1].split(",")
        else:
            strand = 1  # forward oriented
            coords = row[1][1:-1].split(",")
        start, end = int(coords[0]), int(coords[1])

        # Check if this is a tRNA or tmRNA
        if row[0] == "tmRNA":
            ftr = SeqFeature(location=FeatureLocation(start - 1, end),
                             type="tmRNA", strand=strand)
            ftr.qualifiers["gene"] = [""]
            ftr.qualifiers["locus_tag"] = [""]
            tag_peptide = row[-1].rstrip("*")
            ftr.qualifiers["note"] = [f"tag peptide: {tag_peptide}"]
        else:
            ftr = SeqFeature(location=FeatureLocation(start - 1, end),
                             type="tRNA", strand=strand)
            ftr.qualifiers["gene"] = [""]
            ftr.qualifiers["locus_tag"] = [""]
            ftr.qualifiers["note"] = [f"{row[0]}{row[-1]}"]
            if "?" in row[0] or "SeC" in row[0] or "Pyl" in row[0]:
                ftr.qualifiers["product"] = ["tRNA-OTHER"]
            else:
                ftr.qualifiers["product"] = [f"{row[0]}"]

        features.append(ftr)

    # Close the file handle
    aragorn_reader.close()

    return features


def annotate_record(record, tmp_dir, trna=True):
    """
    Uses Prodigal to predict protein-coding genes, and Aragorn to
    predict t(m)RNA genes on bacterial contigs. All resultant features
    are appended directly to the contig's features list.

    :param record: the nucleotide sequence to predict genes on
    :type record: Bio.SeqRecord.SeqRecord
    :param tmp_dir: temporary directory where files can go
    :type tmp_dir: pathlib.Path
    :param trna: don't annotate tRNAs
    :type trna: bool
    """
    # Set up to run Prodigal and Aragorn
    infile = tmp_dir.joinpath(f"{record.id}.fna")

    infile_writer = infile.open("w")
    SeqIO.write(record, infile_writer, "fasta")
    infile_writer.close()

    # Always run Prodigal
    prodigal_out = infile.with_suffix(".faa")
    run_prodigal(infile, prodigal_out, len(record) < META_LENGTH)
    for feature in parse_prodigal(prodigal_out):
        record.features.append(feature)

    # Run Aragorn if trna flag was set
    if trna:
        aragorn_out = infile.with_suffix(".txt")
        run_aragorn(infile, aragorn_out)
        for feature in parse_aragorn(aragorn_out):
            record.features.append(feature)

    # Sort contig features on start position
    record.features.sort(key=lambda x: x.location.start)

    return record


def cleanup_flatfile_records(records):
    """
    Function to clean up and format SeqRecord sequence contigs created
    from imported flat file annotations.

    :param records: imported records
    :type records: list
    """
    for record in records:
        features = list()
        for feature in record.features:
            if feature.type not in CODING_FEATURE_TYPES:
                continue

            if feature.type == "CDS":
                if not feature.qualifiers.get("translation"):
                    dna = feature.extract(record.seq)
                    translation = dna.translate(to_stop=True, table=11)
                    feature.qualifiers["translation"] = [str(translation)]

                feature.qualifiers["product"] = [DEFAULT_PRODUCT]

            features.append(feature)

        record.features = features



def leading_window(window_size, range_size):
    """
    Generator function that yields left/cursor/right index values for
    the prescribed `window_size` ahead of the cursor (inclusive) over
    the range `range_size`.

    :param window_size: size of window (including cursor)
    :type window_size: int
    :param range_size: size of the index range
    :type range_size: int
    """
    for cursor in range(range_size):
        left, right = cursor, cursor + window_size - 1
        yield left, cursor, right


def lagging_window(window_size, range_size):
    """
    Generator function that yields left/cursor/right index values for
    the prescribed `window_size` behind the cursor (inclusive) over
    the range `range_size`.

    :param window_size: size of window (including cursor)
    :type window_size: int
    :param range_size: size of the index range
    :type range_size: int
    """
    for cursor in range(range_size):
        left, right = cursor - window_size + 1, cursor
        yield left, cursor, right


def center_window(window_size, range_size):
    """
    Generator function that yields left/cursor/right index values for
    the prescribed `window_size` around the cursor (inclusive) over the
    range `range_size`.

    NOTE: for even values of `window_size`, windows will be asymmetric,
    with the right side of the window longer than the left side by 1
    index.

    :param window_size: size of window (including cursor)
    :type window_size: int
    :param range_size: size of the index range
    :type range_size: int
    """
    for cursor in range(range_size):
        half = (window_size - 1) // 2
        left, right = cursor - half, cursor + half
        if window_size % 2 == 0:
            right += 1
        yield left, cursor, right



import csv
import json


# READ FUNCTIONS
# -----------------------------------------------------------------------------

def read_gene_index_file(index_file):
    gene_index = {}
    with index_file.open(mode="r") as filehandle:
        lines = filehandle.readlines()
        for line in lines:
            split_line = line.split("\t")
            gene_index[split_line[0]] = {"locus_tag": split_line[1],
                                         "product": split_line[2],
                                         "parent": split_line[3]}

    return gene_index


def read_cluster_table_file(cluster_table_path):
    """Function to read in a cluster table csv

    :param cluster_table_path: Path to a csv table mapping genomes to clusters.
    :type cluster_table: pathlib.Path
    """
    # Use DictReader to read in all rows in the csv as dictionaries,
    # mapping headers to row values
    data_dicts = []
    with cluster_table_path.open(mode="r") as filehandle:
        filereader = csv.DictReader(filehandle)
        for data_dict in filereader:
            data_dicts.append(data_dict)

    # Create a lookup dictionary of genome names to their respective clusters
    record_cluster_map = dict()
    for data_dict in data_dicts:
        record_cluster_map[data_dict["Name"]] = data_dict["Cluster"]

    return record_cluster_map


def read_cluster_index_file(cluster_index):
    with cluster_index.open(mode="r") as filehandle:
        lines = filehandle.readlines()

    cluster_data = []

    line_index = 0
    while True:
        if line_index >= len(lines):
            break

        line = lines[line_index]
        if line.startswith(">"):
            line_index += 1

            if line_index >= len(lines):
                break

            cluster_members = [line.rstrip()
                               for line in lines[line_index].split("\0")]

            cluster_data.append(cluster_members)

        line_index += 1

    return cluster_data


def read_functions_config_file(functions_file):
    accepted_functions = []
    with functions_file.open(mode="r") as filehandle:
        functions_config = json.load(filehandle)

    accepted_functions = functions_config.get("LIKE", list())
    ignored_functions = functions_config.get("NOT LIKE", list())

    return accepted_functions, ignored_functions


# WRITE FUNCTIONS
# -----------------------------------------------------------------------------
def write_cluster_file(clustered_ids, cluster_file):
    with cluster_file.open(mode="w") as cluster_filehandle:
        for cluster_index, record_names in enumerate(clustered_ids):
            cluster_filehandle.write("".join(
                                          [">", str(cluster_index), "\n"]))

            cluster_filehandle.write("".join(
                                          ["\0".join(record_names), "\n"]))


def write_cluster_function_index_file(gene_cluster_data, index_file):
    with index_file.open(mode="w") as filehandle:
        for gene_cluster in gene_cluster_data:
            filehandle.write("\t".join(gene_cluster + ["\n"]))


def write_gene_hex_value_file(gene_hex_value_file, gene_rep_hex_values):
    with gene_hex_value_file.open(mode="wb") as filehandle:
        for hex_value in gene_rep_hex_values:
            filehandle.write(hex_value)
            filehandle.write(b"_")



from depht.functions.subprocess import run_command


def clustalo(fasta_file, aln_out_path, mat_out_path=None, tree_out_path=None,
             outfmt="fasta", infmt="fasta", threads=1, verbose=0):
    """
    Runs Clustal Omega to generate a multiple sequence alignment (MSA)
    and percent identity matrix (PIM) for the indicated file. Infile is
    expected to be in FASTA multiple sequence format. MSA will be in
    Clustal format.
    :param fasta_file: FASTA file containing sequences to be aligned
    :type fasta_file: str or pathlib.Path
    :param aln_out_path: the multiple sequence alignment (MSA) output file
    :type aln_out_path: str or pathlib.Path
    :param mat_out_path: The percent identity matrix (PIM) output file
    :type mat_out_path: str or pathlib.Path
    :param tree_out_path: The alignment guide tree output file
    :type tree_out_path: str or pathlib.Path
    :param outfmt: The file format of the alignment to be exported.
    :type outfmt: str
    :param infmt:  The file format of the sequence file to be read in
    :type infmt: str
    :param threads: number of threads to use for alignment
    :type threads: int
    :param verbose: verbosity level (0-2)
    :type verbose: int
    :return: Returns the sequence alignment path[0] and distance matrix path[1]
    :rtype: tuple
    """
    # Make sure verbose is in range(0,3)
    if verbose <= 0:
        verbose = 0
    elif verbose > 2:
        verbose = 2

    # Build Clustal Omega command that will produce a clustal-formatted
    # alignment output file and percent identity matrix
    command = f"""clustalo -i "{fasta_file}" -o "{aln_out_path}" """ \
              f" --outfmt={outfmt} --infmt={infmt} "\
              f"--force --output-order=tree-order " \
              f"--threads={threads}"

    if mat_out_path is not None:
        command = " ".join([command, (f"""--distmat-out="{mat_out_path}" """
                                      "--full --percent-id")])
    if tree_out_path is not None:
        command = " ".join([command, (
                                f"""--guidetree-out="{tree_out_path}" """)])

    for _ in range(verbose):
        command += " -v"                # Add verbosity to command

    # Run the Clustal Omega command as a subprocess
    run_command(command)

    # Return output files so user can find them easily, whether they specified
    # the filenames or not
    return (aln_out_path, mat_out_path)
